@@ -1,6 +1,5 @@
 # -*- coding: utf-8 -*-
-"""Test bre lib."""
-from __future__ import unicode_literals
+"""Test `bre` lib."""
 import unittest
 from backrefs import bre
 import re
@@ -10,21 +9,49 @@ import sre_constants
 import random
 from backrefs import _bre_parse
 import copy
+import warnings
 
-PY3 = (3, 0) <= sys.version_info < (4, 0)
-PY34_PLUS = (3, 4) <= sys.version_info
 PY36_PLUS = (3, 6) <= sys.version_info
 PY37_PLUS = (3, 7) <= sys.version_info
-
-if PY3:
-    binary_type = bytes  # noqa
-else:
-    binary_type = str  # noqa
+PY39_PLUS = (3, 9) <= sys.version_info
 
 
 class TestSearchTemplate(unittest.TestCase):
     """Search template tests."""
 
+    def test_posix_unicode_and_range(self):
+        """POSIX and Unicode classes should not be part of a range."""
+
+        self.assertTrue(bre.match(r'[[:upper:]-[:digit:]]', '-') is not None)
+        self.assertTrue(bre.match(r'[[:upper:]-[:digit:]]', 'Z') is not None)
+        self.assertTrue(bre.match(r'[[:upper:]-[:digit:]]', '0') is not None)
+        self.assertTrue(bre.match(r'[[:upper:]-[:digit:]]', '5') is not None)
+
+        self.assertTrue(bre.match(r'[[:digit:]-[:upper:]]', 'A') is not None)
+        self.assertTrue(bre.match(r'[[:digit:]-[:upper:]]', '9') is not None)
+        self.assertTrue(bre.match(r'[[:digit:]-[:upper:]]', '-') is not None)
+        self.assertTrue(bre.match(r'[[:digit:]-[:upper:]]', 'D') is not None)
+
+        self.assertTrue(bre.match(r'[0-[:upper:]]', '-') is not None)
+        self.assertTrue(bre.match(r'[0-[:upper:]]', '0') is not None)
+        self.assertTrue(bre.match(r'[0-[:upper:]]', 'A') is not None)
+        self.assertTrue(bre.match(r'[0-[:upper:]]', 'D') is not None)
+
+        self.assertTrue(bre.match(r'[[:digit:]-Z]', 'Z') is not None)
+        self.assertTrue(bre.match(r'[[:digit:]-Z]', '9') is not None)
+        self.assertTrue(bre.match(r'[[:digit:]-Z]', '-') is not None)
+        self.assertTrue(bre.match(r'[[:digit:]-Z]', '5') is not None)
+
+        self.assertTrue(bre.match(r'[\p{digit}-Z]', 'Z') is not None)
+        self.assertTrue(bre.match(r'[\p{digit}-Z]', '9') is not None)
+        self.assertTrue(bre.match(r'[\p{digit}-Z]', '-') is not None)
+        self.assertTrue(bre.match(r'[\p{digit}-Z]', '5') is not None)
+
+        self.assertTrue(bre.match(r'[0-\c]', '-') is not None)
+        self.assertTrue(bre.match(r'[0-\c]', '0') is not None)
+        self.assertTrue(bre.match(r'[0-\c]', 'A') is not None)
+        self.assertTrue(bre.match(r'[0-\c]', 'D') is not None)
+
     def test_non_raw_string_unicode(self):
         """Test non raw string Unicode notation."""
 
@@ -60,13 +87,15 @@ class TestSearchTemplate(unittest.TestCase):
         p1 = bre.compile('test')
         p2 = bre.compile('test')
         p3 = bre.compile('test', bre.X)
+        p4 = bre.compile(b'test')
 
         self.assertTrue(p1 == p2)
         self.assertTrue(p1 != p3)
+        self.assertTrue(p1 != p4)
 
-        p4 = copy.copy(p1)
-        self.assertTrue(p1 == p4)
-        self.assertTrue(p4 in {p1})
+        p5 = copy.copy(p1)
+        self.assertTrue(p1 == p5)
+        self.assertTrue(p5 in {p1})
 
     def test_not_flags(self):
         """Test invalid flags."""
@@ -182,9 +211,8 @@ class TestSearchTemplate(unittest.TestCase):
     def test_infinite_loop_catch(self):
         """Test infinite loop catch."""
 
-        if PY3:
-            with pytest.raises(_bre_parse.LoopException):
-                bre.compile_search(r'((?a)(?u))')
+        with pytest.raises(_bre_parse.LoopException):
+            bre.compile_search(r'((?a)(?u))')
         if PY36_PLUS:
             with pytest.raises(_bre_parse.LoopException):
                 bre.compile_search(r'(?-x:(?x))', re.VERBOSE)
@@ -426,12 +454,9 @@ class TestSearchTemplate(unittest.TestCase):
 
     def test_posix_in_group_ascii(self):
         """Test posix in a group for ASCII."""
-        if PY3:
-            pattern = bre.compile_search(r'Test [[:graph:]]', re.ASCII)
-            pattern2 = bre.compile_search('Test [\u0021-\u007E]', re.ASCII)
-        else:
-            pattern = bre.compile_search(r'Test [[:graph:]]')
-            pattern2 = bre.compile_search(r'Test [\u0021-\u007E]')
+
+        pattern = bre.compile_search(r'Test [[:graph:]]', re.ASCII)
+        pattern2 = bre.compile_search('Test [!-\\~]', re.ASCII)
         self.assertEqual(pattern.pattern, pattern2.pattern)
 
     def test_not_posix_at_start_group(self):
@@ -654,7 +679,7 @@ class TestSearchTemplate(unittest.TestCase):
     def test_unicode_shorthand_ascii_only(self):
         """Ensure that when the Unicode flag is not used, only ASCII properties are used."""
 
-        flags = bre.ASCII if PY3 else 0
+        flags = bre.ASCII
         pattern = bre.compile_search(r'ex\lmple', flags)
         m = pattern.match('exámple')
         self.assertTrue(m is None)
@@ -756,14 +781,14 @@ class TestSearchTemplate(unittest.TestCase):
         m = pattern.match(r'P')
         self.assertTrue(m is None)
 
-    def test_binary_property(self):
-        r"""Binary patterns should match `\p` references."""
+    def test_bytes_property(self):
+        r"""Byte patterns should match `\p` references."""
 
         pattern = bre.compile_search(br'EX\p{Lu}MPLE')
         m = pattern.match(br'EXAMPLE')
         self.assertTrue(m is not None)
 
-    def test_binary_property_no_value(self):
+    def test_bytes_property_no_value(self):
         """Test when a property returns nothing in byte string."""
 
         pattern = bre.compile_search(br'EX\p{OtherAlphabetic}MPLE')
@@ -869,62 +894,37 @@ class TestSearchTemplate(unittest.TestCase):
     def test_unicode_string_flag(self):
         """Test finding Unicode/ASCII string flag."""
 
-        if PY3:
-            template = _bre_parse._SearchParser(r'Testing for (?ia) ASCII flag.', False, None)
-            template.parse()
-            self.assertFalse(template.unicode)
-        else:
-            template = _bre_parse._SearchParser(r'Testing for (?iu) Unicode flag.', False, None)
-            template.parse()
-            self.assertTrue(template.unicode)
+        template = _bre_parse._SearchParser(r'Testing for (?ia) ASCII flag.', False, None)
+        template.parse()
+        self.assertFalse(template.unicode)
 
     def test_unicode_string_flag_in_group(self):
         """Test ignoring Unicode/ASCII string flag in group."""
 
-        if PY3:
-            template = _bre_parse._SearchParser(r'Testing for [(?ia)] ASCII flag.', False, None)
-            template.parse()
-            self.assertTrue(template.unicode)
-        else:
-            template = _bre_parse._SearchParser(r'Testing for [(?iu)] Unicode flag.', False, None)
-            template.parse()
-            self.assertFalse(template.unicode)
+        template = _bre_parse._SearchParser(r'Testing for [(?ia)] ASCII flag.', False, None)
+        template.parse()
+        self.assertTrue(template.unicode)
 
     def test_unicode_string_flag_escaped(self):
         """Test ignoring Unicode/ASCII string flag in group."""
 
-        if PY3:
-            template = _bre_parse._SearchParser(r'Testing for \(?ia) ASCII flag.', False, None)
-            template.parse()
-            self.assertTrue(template.unicode)
-        else:
-            template = _bre_parse._SearchParser(r'Testing for \(?iu) Unicode flag.', False, None)
-            template.parse()
-            self.assertFalse(template.unicode)
+        template = _bre_parse._SearchParser(r'Testing for \(?ia) ASCII flag.', False, None)
+        template.parse()
+        self.assertTrue(template.unicode)
 
     def test_unicode_string_flag_unescaped(self):
         """Test unescaped Unicode string flag."""
 
-        if PY3:
-            template = _bre_parse._SearchParser(r'Testing for \\(?ia) ASCII flag.', False, None)
-            template.parse()
-            self.assertFalse(template.unicode)
-        else:
-            template = _bre_parse._SearchParser(r'Testing for \\(?iu) Unicode flag.', False, None)
-            template.parse()
-            self.assertTrue(template.unicode)
+        template = _bre_parse._SearchParser(r'Testing for \\(?ia) ASCII flag.', False, None)
+        template.parse()
+        self.assertFalse(template.unicode)
 
     def test_unicode_string_flag_escaped_deep(self):
         """Test deep escaped Unicode flag."""
 
-        if PY3:
-            template = _bre_parse._SearchParser(r'Testing for \\\(?ia) ASCII flag.', False, None)
-            template.parse()
-            self.assertTrue(template.unicode)
-        else:
-            template = _bre_parse._SearchParser(r'Testing for \\\(?iu) Unicode flag.', False, None)
-            template.parse()
-            self.assertFalse(template.unicode)
+        template = _bre_parse._SearchParser(r'Testing for \\\(?ia) ASCII flag.', False, None)
+        template.parse()
+        self.assertTrue(template.unicode)
 
     def test_verbose_comment_no_nl(self):
         """Test verbose comment with no newline."""
@@ -964,6 +964,12 @@ class TestSearchTemplate(unittest.TestCase):
         m = pattern2.match('test')
         self.assertTrue(m is not None)
 
+    @unittest.skipUnless(PY39_PLUS, "Python 3.9 required")
+    def test_emoji_property(self):
+        """Test new emoji properties."""
+
+        self.assertEqual(len(bre.findall(r'\p{emoji}', '\U0001F600 test \U0001F600')), 2)
+
 
 class TestReplaceTemplate(unittest.TestCase):
     """Test replace template."""
@@ -973,26 +979,29 @@ class TestReplaceTemplate(unittest.TestCase):
 
         p1 = bre.compile('(test)')
         p2 = bre.compile('(test)')
+        p3 = bre.compile(b'(test)')
         r1 = p1.compile(r'\1')
         r2 = p1.compile(r'\1')
         r3 = p2.compile(r'\1')
         r4 = p2.compile(r'\1', bre.FORMAT)
+        r5 = p3.compile(br'\1')
 
         self.assertTrue(r1 == r2)
         self.assertTrue(r2 == r3)
         self.assertTrue(r1 != r4)
+        self.assertTrue(r1 != r5)
 
-        r5 = copy.copy(r1)
-        self.assertTrue(r1 == r5)
-        self.assertTrue(r5 in {r1})
+        r6 = copy.copy(r1)
+        self.assertTrue(r1 == r6)
+        self.assertTrue(r6 in {r1})
 
     def test_format_failures(self):
         """Test format parsing failures."""
 
-        with pytest.raises(SyntaxError):
+        with pytest.raises(sre_constants.error if PY36_PLUS else IndexError):
             bre.subf('test', r'{1.}', 'test', bre.FORMAT)
 
-        with pytest.raises(SyntaxError):
+        with pytest.raises(IndexError):
             bre.subf('test', r'{a.}', 'test', bre.FORMAT)
 
         with pytest.raises(SyntaxError):
@@ -1010,6 +1019,57 @@ class TestReplaceTemplate(unittest.TestCase):
         with pytest.raises(SyntaxError):
             bre.subf('test', r'test { test', 'test', bre.FORMAT)
 
+        with pytest.raises(sre_constants.error if PY36_PLUS else IndexError):
+            bre.subf(b'test', br'{1.}', b'test', bre.FORMAT)
+
+        with pytest.raises(IndexError):
+            bre.subf(b'test', br'{a.}', b'test', bre.FORMAT)
+
+        with pytest.raises(SyntaxError):
+            bre.subf(b'test', br'{1[}', b'test', bre.FORMAT)
+
+        with pytest.raises(SyntaxError):
+            bre.subf(b'test', br'{a[}', b'test', bre.FORMAT)
+
+        with pytest.raises(SyntaxError):
+            bre.subf(b'test', br'test } test', b'test', bre.FORMAT)
+
+        with pytest.raises(SyntaxError):
+            bre.subf(b'test', br'test {test', b'test', bre.FORMAT)
+
+        with pytest.raises(SyntaxError):
+            bre.subf(b'test', br'test { test', b'test', bre.FORMAT)
+
+        with pytest.raises(TypeError):
+            bre.subf(b'test', br'{[test]}', b'test', bre.FORMAT)
+
+    def test_format_conversions(self):
+        """Test string format conversion paths."""
+
+        self.assertTrue(bre.subf('test', r'{0.index}', 'test').startswith('<built-in method'))
+        self.assertEqual(bre.subf('test', r'{0.__class__.__name__}', 'test'), 'str')
+        self.assertTrue(bre.subf('test', r'{0.index!s}', 'test').startswith('<built-in method'))
+        self.assertEqual(bre.subf('test', r'{0.__class__.__name__!s}', 'test'), 'str')
+        self.assertTrue(bre.subf('test', r'{0.index!a}', 'test').startswith('<built-in method'))
+
+        self.assertTrue(bre.subf(b'test', br'{0.index}', b'test').startswith(b'<built-in method'))
+        self.assertEqual(bre.subf(b'test', br'{0.__class__.__name__}', b'test'), b'bytes')
+        self.assertTrue(bre.subf(b'test', br'{0.index!s}', b'test').startswith(b'<built-in method'))
+        self.assertEqual(bre.subf(b'test', br'{0.__class__.__name__!s}', b'test'), b'bytes')
+        self.assertTrue(bre.subf('test', r'{0.index!a}', 'test').startswith('<built-in method'))
+
+    def test_incompatible_strings(self):
+        """Test incompatible string types."""
+
+        with pytest.raises(TypeError):
+            bre.compile('test').compile(b'test')
+
+        p1 = bre.compile('test')
+        repl = bre.compile(b'test').compile(b'other')
+        m = p1.match('test')
+        with pytest.raises(TypeError):
+            repl(m)
+
     def test_named_unicode_failures(self):
         """Test named Unicode failures."""
 
@@ -1060,14 +1120,10 @@ class TestReplaceTemplate(unittest.TestCase):
     def test_unicode_narrow_value(self):
         """Test Unicode narrow value."""
 
-        # Python 3 does not apply Unicode syntax in raw string,
-        # but Python 2 does.  So Python 2 will turn the following
-        # into a group and fail. There isn't really a workaround for this.
-        if PY3:
-            pattern = re.compile(r"(some)(.+?)(pattern)(!)")
-            expand = bre.compile_replace(pattern, r'\u005cg')
-            results = expand(pattern.match('some test pattern!'))
-            self.assertEqual('\g', results)
+        pattern = re.compile(r"(some)(.+?)(pattern)(!)")
+        expand = bre.compile_replace(pattern, r'\u005cg')
+        results = expand(pattern.match('some test pattern!'))
+        self.assertEqual(r'\g', results)
 
     def test_unexpected_end(self):
         """Test cases where there is an unexpected end to the replace string."""
@@ -1082,6 +1138,56 @@ class TestReplaceTemplate(unittest.TestCase):
         with pytest.raises(sre_constants.error):
             _bre_parse._ReplaceParser().parse(pattern, '\\1\\')
 
+    def test_line_break(self):
+        r"""Test line break \R."""
+
+        self.assertEqual(
+            bre.sub(r"\R", ' ', 'line\r\nline\nline\r'),
+            'line line line '
+        )
+
+    def test_bytes_line_break(self):
+        r"""Test bytes line break \R."""
+
+        self.assertEqual(
+            bre.sub(br"\R", b' ', b'line\r\nline\nline\r'),
+            b'line line line '
+        )
+
+    def test_line_break_in_group(self):
+        """Test that line break in group matches a normal R."""
+
+        if PY36_PLUS:
+            with pytest.raises(sre_constants.error):
+                bre.sub(r"[\R]", 'l', 'Rine\r\nRine\nRine\r')
+        else:
+            self.assertEqual(
+                bre.sub(r"[\R]", 'l', 'Rine\r\nRine\nRine\r'),
+                'line\r\nline\nline\r'
+            )
+
+    def test_grapheme_cluster(self):
+        """Test simple grapheme cluster."""
+
+        self.assertEqual(
+            bre.compile(r'\X').pattern,
+            bre.compile(r'(?:\PM\pM*(?!\pM))').pattern
+        )
+
+        self.assertEqual(bre.match(r"\X", "\xE0").span(), (0, 1))
+        self.assertEqual(bre.match(r"\X", "a\u0300").span(), (0, 2))
+
+        self.assertEqual(
+            bre.findall(r"\X", "a\xE0a\u0300e\xE9e\u0301"),
+            ['a', '\xe0', 'a\u0300', 'e', '\xe9', 'e\u0301']
+        )
+        self.assertEqual(
+            bre.findall(r"\X{3}", "a\xE0a\u0300e\xE9e\u0301"),
+            ['a\xe0a\u0300', 'e\xe9e\u0301']
+        )
+        # `self.assertEqual(regex.findall(r"\X", "\r\r\n\u0301A\u0301"), ['\r', '\r\n', '\u0301', 'A\u0301'])`
+        self.assertEqual(bre.search(r'\X$', 'ab\u2103').group(), '\u2103')
+
     def test_replace_unicode_name_ascii_range(self):
         """Test replacing Unicode names in the ASCII range."""
 
@@ -1188,7 +1294,7 @@ class TestReplaceTemplate(unittest.TestCase):
         expand = bre.compile_replace(pattern, r'\1\c\l\2\3')
         results = expand(pattern.match(text))
 
-        self.assertEqual('This is a test for Stacking!', results)
+        self.assertEqual('This is a test for stacking!', results)
 
     def test_span_stacked_case(self):
         """Test stacked casing of non-spans in and out of a span."""
@@ -1198,7 +1304,7 @@ class TestReplaceTemplate(unittest.TestCase):
         expand = bre.compile_replace(pattern, r'\1\c\L\l\2\E\3')
         results = expand(pattern.match(text))
 
-        self.assertEqual('This is a test for Stacking!', results)
+        self.assertEqual('This is a test for stacking!', results)
 
     def test_single_case_followed_by_bslash(self):
         """Test single backslash following a single case reference."""
@@ -1228,7 +1334,7 @@ class TestReplaceTemplate(unittest.TestCase):
         expand = bre.compile_replace(pattern, r'Test \l\Cstacked\E\3')
         results = expand(pattern.match(text))
 
-        self.assertEqual('Test sTACKED!', results)
+        self.assertEqual('Test STACKED!', results)
 
     def test_extraneous_end_char(self):
         """Test for extraneous end characters."""
@@ -1318,7 +1424,7 @@ class TestReplaceTemplate(unittest.TestCase):
         expand = bre.compile_replace(pattern, r'\1\C\C\2\E\3\E')
         results = expand(pattern.match(text))
 
-        self.assertEqual('This is a complex UPPERCASE TEST!', results)
+        self.assertEqual('This is a complex UPPERCASE test!', results)
 
     def test_lower_after_lower(self):
         """Test lowercase after lowercase."""
@@ -1338,7 +1444,7 @@ class TestReplaceTemplate(unittest.TestCase):
         expand = bre.compile_replace(pattern, r'\1\L\L\2\E\3\E')
         results = expand(pattern.match(text))
 
-        self.assertEqual('This is a complex lowercase test!', results)
+        self.assertEqual('This is a complex lowercase TEST!', results)
 
     def test_span_upper_after_lower(self):
         """Test lowercase followed by uppercase span."""
@@ -1348,7 +1454,7 @@ class TestReplaceTemplate(unittest.TestCase):
         expand = bre.compile_replace(pattern, r'\1\l\C\2\E\3')
         results = expand(pattern.match(text))
 
-        self.assertEqual('This is a complex uPPERCASE test!', results)
+        self.assertEqual('This is a complex UPPERCASE test!', results)
 
     def test_span_lower_after_upper(self):
         """Test uppercase followed by lowercase span."""
@@ -1358,7 +1464,7 @@ class TestReplaceTemplate(unittest.TestCase):
         expand = bre.compile_replace(pattern, r'\1\c\L\2\E\3')
         results = expand(pattern.match(text))
 
-        self.assertEqual('This is a complex Lowercase test!', results)
+        self.assertEqual('This is a complex lowercase test!', results)
 
     def test_span_upper_around_lower(self):
         """Test uppercase span around a lowercase."""
@@ -1464,8 +1570,8 @@ class TestReplaceTemplate(unittest.TestCase):
         self.assertEqual(results2, results)
         self.assertEqual('\t \\t \\\t \\\\t \\\\\t', results)
 
-    def test_binary_normal_escaping(self):
-        """Test binary normal escaped slash."""
+    def test_bytes_normal_escaping(self):
+        """Test bytes normal escaped slash."""
 
         text = b"This is a test of normal escaping!"
         pattern = re.compile(br"(.+)")
@@ -1525,7 +1631,7 @@ class TestReplaceTemplate(unittest.TestCase):
         # Use text pattern directly.
         expand = bre.compile_replace(pattern, r'\l\C\g<first>\l\g<second>\L\c\g<third>\E\g<fourth>\E\5')
         results = expand(pattern.match(text))
-        self.assertEqual('tHIS iS A TEST FOR Named capture GROUPS!', results)
+        self.assertEqual('THIS iS A TEST FOR Named capture groups!', results)
 
     def test_mixed_groups2(self):
         """Test mix of upper and lower case with group indexes and a string replace pattern."""
@@ -1537,7 +1643,7 @@ class TestReplaceTemplate(unittest.TestCase):
         # This will pass because we do not need to resolve named groups.
         expand = bre.compile_replace(pattern, r'\l\C\g<1>\l\g<2>\L\c\g<3>\E\g<4>\E\5')
         results = expand(pattern.match(text))
-        self.assertEqual('tHIS iS A TEST FOR Named capture GROUPS!', results)
+        self.assertEqual('THIS iS A TEST FOR Named capture groups!', results)
 
     def test_mixed_groups3(self):
         """Test mix of upper and lower case with named groups and a compiled replace pattern."""
@@ -1549,7 +1655,7 @@ class TestReplaceTemplate(unittest.TestCase):
         # Now using compiled pattern, we can use named groups in replace template.
         expand = bre.compile_replace(pattern, r'\l\C\g<first>\l\g<second>\L\c\g<third>\E\g<fourth>\E\5')
         results = expand(pattern.match(text))
-        self.assertEqual('tHIS iS A TEST FOR Named capture GROUPS!', results)
+        self.assertEqual('THIS iS A TEST FOR Named capture groups!', results)
 
     def test_as_replace_function(self):
         """Test that replace can be used as a replace function."""
@@ -1563,16 +1669,16 @@ class TestReplaceTemplate(unittest.TestCase):
         self.assertEqual(result, "This is awesome!  Here we go!  This is awesome!  Here we go!")
         self.assertEqual(count, 2)
 
-    def test_binary_replace(self):
-        """Test that binary regex result is a binary string."""
+    def test_bytes_replace(self):
+        """Test that bytes regex result is a bytes string."""
 
-        text = b"This is some binary text!"
-        pattern = bre.compile_search(br"This is (some binary text)!")
+        text = b"This is some bytes text!"
+        pattern = bre.compile_search(br"This is (some bytes text)!")
         expand = bre.compile_replace(pattern, br'\C\1\E')
         m = pattern.match(text)
         result = expand(m)
-        self.assertEqual(result, b"SOME BINARY TEXT")
-        self.assertTrue(isinstance(result, binary_type))
+        self.assertEqual(result, b"SOME BYTES TEXT")
+        self.assertTrue(isinstance(result, bytes))
 
     def test_template_replace(self):
         """Test replace by passing in replace function."""
@@ -1598,7 +1704,7 @@ class TestReplaceTemplate(unittest.TestCase):
         expand = bre.compile_replace(pattern, r'\l\C\g<0001>\l\g<02>\L\c\g<03>\E\g<004>\E\5\n\C\g<000>\E')
         results = expand(pattern.match(text))
         self.assertEqual(
-            'tHIS iS A TEST FOR Numeric capture GROUPS!\nTHIS IS A TEST FOR NUMERIC CAPTURE GROUPS!',
+            'THIS iS A TEST FOR Numeric capture groups!\nTHIS IS A TEST FOR NUMERIC CAPTURE GROUPS!',
             results
         )
 
@@ -1609,10 +1715,21 @@ class TestReplaceTemplate(unittest.TestCase):
         text_pattern = r"(this )(.+?)(numeric capture )(groups)(!)"
         pattern = re.compile(text_pattern)
 
-        expand = bre.compile_replace(pattern, r'\l\C{0001}\l{02}\L\c{03}\E{004}\E{5}\n\C{000}\E', bre.FORMAT)
-        results = expand(pattern.match(text))
+        expandf = bre.compile_replace(pattern, r'\l\C{0001}\l{02}\L\c{03}\E{004}\E{5}\n\C{000}\E', bre.FORMAT)
+        results = expandf(pattern.match(text))
         self.assertEqual(
-            'tHIS iS A TEST FOR Numeric capture GROUPS!\nTHIS IS A TEST FOR NUMERIC CAPTURE GROUPS!',
+            'THIS iS A TEST FOR Numeric capture groups!\nTHIS IS A TEST FOR NUMERIC CAPTURE GROUPS!',
+            results
+        )
+
+        text = b"this is a test for numeric capture groups!"
+        text_pattern = br"(this )(.+?)(numeric capture )(groups)(!)"
+        pattern = re.compile(text_pattern)
+
+        expandf = bre.compile_replace(pattern, br'\l\C{0001}\l{02}\L\c{03}\E{004}\E{5}\n\C{000}\E', bre.FORMAT)
+        results = expandf(pattern.match(text))
+        self.assertEqual(
+            b'THIS iS A TEST FOR Numeric capture groups!\nTHIS IS A TEST FOR NUMERIC CAPTURE GROUPS!',
             results
         )
 
@@ -1623,12 +1740,25 @@ class TestReplaceTemplate(unittest.TestCase):
         text_pattern = r"(this )(.+?)(format capture )(groups)(!)"
         pattern = re.compile(text_pattern)
 
-        expand = bre.compile_replace(
+        expandf = bre.compile_replace(
             pattern, r'\l\C{{0001}}\l{{{02}}}\L\c{03}\E{004}\E{5}\n\C{000}\E', bre.FORMAT
         )
-        results = expand(pattern.match(text))
+        results = expandf(pattern.match(text))
         self.assertEqual(
-            '{0001}{IS A TEST FOR }Format capture GROUPS!\nTHIS IS A TEST FOR FORMAT CAPTURE GROUPS!',
+            '{0001}{IS A TEST FOR }Format capture groups!\nTHIS IS A TEST FOR FORMAT CAPTURE GROUPS!',
+            results
+        )
+
+        text = b"this is a test for format capture groups!"
+        text_pattern = br"(this )(.+?)(format capture )(groups)(!)"
+        pattern = re.compile(text_pattern)
+
+        expandf = bre.compile_replace(
+            pattern, br'\l\C{{0001}}\l{{{02}}}\L\c{03}\E{004}\E{5}\n\C{000}\E', bre.FORMAT
+        )
+        results = expandf(pattern.match(text))
+        self.assertEqual(
+            b'{0001}{IS A TEST FOR }Format capture groups!\nTHIS IS A TEST FOR FORMAT CAPTURE GROUPS!',
             results
         )
 
@@ -1639,12 +1769,25 @@ class TestReplaceTemplate(unittest.TestCase):
         text_pattern = r"(this )(.+?)(format capture )(groups)(!)"
         pattern = re.compile(text_pattern)
 
-        expand = bre.compile_replace(
+        expandf = bre.compile_replace(
             pattern, r'\C{}\E\n\l\C{}\l{}\L\c{}\E{}\E{}{{}}', bre.FORMAT
         )
-        results = expand(pattern.match(text))
+        results = expandf(pattern.match(text))
         self.assertEqual(
-            'THIS IS A TEST FOR FORMAT CAPTURE GROUPS!\ntHIS iS A TEST FOR Format capture GROUPS!{}',
+            'THIS IS A TEST FOR FORMAT CAPTURE GROUPS!\nTHIS iS A TEST FOR Format capture groups!{}',
+            results
+        )
+
+        text = b"this is a test for format capture groups!"
+        text_pattern = br"(this )(.+?)(format capture )(groups)(!)"
+        pattern = re.compile(text_pattern)
+
+        expandf = bre.compile_replace(
+            pattern, br'\C{}\E\n\l\C{}\l{}\L\c{}\E{}\E{}{{}}', bre.FORMAT
+        )
+        results = expandf(pattern.match(text))
+        self.assertEqual(
+            b'THIS IS A TEST FOR FORMAT CAPTURE GROUPS!\nTHIS iS A TEST FOR Format capture groups!{}',
             results
         )
 
@@ -1668,7 +1811,7 @@ class TestReplaceTemplate(unittest.TestCase):
         """Test format auto capture indexing."""
 
         text = "abababab"
-        text_pattern = r"(\w)+"
+        text_pattern = r"(\w{2})+"
         pattern = re.compile(text_pattern)
 
         expand = bre.compile_replace(
@@ -1676,7 +1819,7 @@ class TestReplaceTemplate(unittest.TestCase):
         )
         results = expand(pattern.match(text))
         self.assertEqual(
-            'ababababb',
+            'ba',
             results
         )
 
@@ -1696,7 +1839,7 @@ class TestReplaceTemplate(unittest.TestCase):
             results
         )
 
-    def test_binary_format_capture_bases(self):
+    def test_bytes_format_capture_bases(self):
         """Test capture bases."""
 
         text = b"abababab"
@@ -1706,11 +1849,9 @@ class TestReplaceTemplate(unittest.TestCase):
         expand = bre.compile_replace(
             pattern, br'{1[-0x1]}{1[-0o1]}{1[-0b1]}', bre.FORMAT
         )
+
         results = expand(pattern.match(text))
-        self.assertEqual(
-            b'bbb',
-            results
-        )
+        self.assertEqual(b'989898', results)
 
     def test_format_escapes(self):
         """Test format escapes."""
@@ -1736,7 +1877,7 @@ class TestReplaceTemplate(unittest.TestCase):
         pattern = re.compile(text_pattern)
 
         expand = bre.compile_replace(
-            pattern, r'\{1[-1]}\\{1[-1]}', bre.FORMAT
+            pattern, r'\{1}\\{1}', bre.FORMAT
         )
         results = expand(pattern.match(text))
         self.assertEqual(
@@ -1751,6 +1892,99 @@ class TestReplaceTemplate(unittest.TestCase):
         result = pattern.sub('\\C\\U00000070\\U0001F360\\E', 'Test')
         self.assertEqual(result, 'P\U0001F360')
 
+    def test_format_inter_escape(self):
+        """Test escaped characters inside format group."""
+
+        self.assertEqual(
+            bre.subf(
+                r'(Te)(st)(ing)(!)',
+                r'\x7b1\x7d-\u007b2\u007d-\U0000007b3\U0000007d-\N{LEFT CURLY BRACKET}4\N{RIGHT CURLY BRACKET}',
+                'Testing!'
+            ),
+            "Te-st-ing-!"
+        )
+        self.assertEqual(
+            bre.subf(
+                r'(Te)(st)(ing)(!)',
+                r'\1731\175-{2:\\^6}',
+                'Testing!'
+            ),
+            "Te-\\\\st\\\\"
+        )
+
+        with pytest.raises(SyntaxError):
+            bre.subf(r'(test)', r'{\g}', 'test')
+
+        with pytest.raises(ValueError):
+            bre.subf(br'(test)', br'{\777}', b'test')
+
+    def test_format_features(self):
+        """Test format features."""
+
+        pattern = bre.compile(r'(Te)(st)(?P<group>ing)')
+        self.assertEqual(pattern.subf(r'{.__class__.__name__}', 'Testing'), 'str')
+        self.assertEqual(pattern.subf(r'{1:<30}', 'Testing'), 'Te                            ')
+        self.assertEqual(pattern.subf(r'{1:30}', 'Testing'), 'Te                            ')
+        self.assertEqual(pattern.subf(r'{1:>30}', 'Testing'), '                            Te')
+        self.assertEqual(pattern.subf(r'{1:^30}', 'Testing'), '              Te              ')
+        self.assertEqual(pattern.subf(r'{1:*^30}', 'Testing'), '**************Te**************')
+        self.assertEqual(pattern.subf(r'{1:^030}', 'Testing'), '00000000000000Te00000000000000')
+        self.assertEqual(pattern.subf(r'{1:^^30}', 'Testing'), '^^^^^^^^^^^^^^Te^^^^^^^^^^^^^^')
+        self.assertEqual(pattern.subf(r'{1:1^30}', 'Testing'), '11111111111111Te11111111111111')
+        self.assertEqual(pattern.subf(r'{1:<30s}', 'Testing'), 'Te                            ')
+        self.assertEqual(pattern.subf(r'{1:s}', 'Testing'), 'Te')
+        self.assertEqual(pattern.subf(r'{2!r}', 'Testing'), "'st'")
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(r'{2!x}', 'Testing')
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(r'{2$}', 'Testing')
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(r'{2$}', 'Testing')
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(r'{a$}', 'Testing')
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(r'{:ss}', 'Testing')
+
+        with pytest.raises(ValueError):
+            pattern.subf(r'{:030}', 'Testing')
+
+        pattern = bre.compile(br'(Te)(st)(?P<group>ing)')
+        self.assertEqual(pattern.subf(br'{.__class__.__name__}', b'Testing'), b'bytes')
+        self.assertEqual(pattern.subf(br'{1:<30}', b'Testing'), b'Te                            ')
+        self.assertEqual(pattern.subf(br'{1:30}', b'Testing'), b'Te                            ')
+        self.assertEqual(pattern.subf(br'{1:>30}', b'Testing'), b'                            Te')
+        self.assertEqual(pattern.subf(br'{1:^30}', b'Testing'), b'              Te              ')
+        self.assertEqual(pattern.subf(br'{1:*^30}', b'Testing'), b'**************Te**************')
+        self.assertEqual(pattern.subf(br'{1:^030}', b'Testing'), b'00000000000000Te00000000000000')
+        self.assertEqual(pattern.subf(br'{1:^^30}', b'Testing'), b'^^^^^^^^^^^^^^Te^^^^^^^^^^^^^^')
+        self.assertEqual(pattern.subf(br'{1:1^30}', b'Testing'), b'11111111111111Te11111111111111')
+        self.assertEqual(pattern.subf(br'{1:<30s}', b'Testing'), b'Te                            ')
+        self.assertEqual(pattern.subf(br'{1:s}', b'Testing'), b'Te')
+        self.assertEqual(pattern.subf(br'{2!r}', b'Testing'), b"b'st'")
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(br'{2!x}', b'Testing')
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(br'{2$}', b'Testing')
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(br'{2$}', b'Testing')
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(br'{a$}', b'Testing')
+
+        with pytest.raises(SyntaxError):
+            pattern.subf(br'{:ss}', b'Testing')
+
+        with pytest.raises(ValueError):
+            pattern.subf(br'{:030}', b'Testing')
+
     def test_dont_case_special_refs(self):
         """Test that we don't case Unicode and bytes tokens, but case the character."""
 
@@ -1777,7 +2011,7 @@ class TestReplaceTemplate(unittest.TestCase):
         # Python 3.7 will choke on \U and \u as invalid escapes, so
         # don't bother running these tests there.
         if not PY37_PLUS:
-            # Binary doesn't care about Unicode, but should evaluate bytes
+            # Bytes doesn't care about Unicode, but should evaluate bytes
             pattern = re.compile(b'Test')
             expand = bre.compile_replace(pattern, br'\C\u0109\n\x77\E\l\x57\c\u0109')
             results = expand(pattern.match(b'Test'))
@@ -1843,22 +2077,6 @@ class TestReplaceTemplate(unittest.TestCase):
 class TestExceptions(unittest.TestCase):
     """Test Exceptions."""
 
-    # def test_incomplete_replace_narrow_unicode(self):
-    #     """Test incomplete replace of narrow Unicode."""
-
-    #     p = bre.compile_search(r'test')
-    #     with self.assertRaises(SyntaxError) as e:
-    #         bre.compile_replace(p, r'Replace \u fail!')
-    #     self.assertTrue(str(e), 'Format for Unicode is \\uXXXX!')
-
-    # def test_incomplete_replace_wide_unicode(self):
-    #     """Test incomplete replace wide Unicode."""
-
-    #     p = bre.compile_search(r'test')
-    #     with self.assertRaises(SyntaxError) as e:
-    #         bre.compile_replace(p, r'Replace \U fail!')
-    #     self.assertTrue(str(e), 'Format for wide Unicode is \\UXXXXXXXX!')
-
     def test_immutable(self):
         """Test immutable object."""
 
@@ -1889,16 +2107,16 @@ class TestExceptions(unittest.TestCase):
 
         self.assertEqual(str(e.value), 'Invalid POSIX property!')
 
-    def test_bad_binary(self):
-        """Test bad binary."""
+    def test_bad_bytes(self):
+        """Test bad bytes."""
 
         with pytest.raises(ValueError) as e:
-            bre.compile_search(r'\p{bad_binary:n}', re.UNICODE)
+            bre.compile_search(r'\p{bad_bytes:n}', re.UNICODE)
 
         self.assertEqual(str(e.value), 'Invalid Unicode property!')
 
         with pytest.raises(ValueError) as e:
-            bre.compile_search(r'\p{bad_binary:y}', re.UNICODE)
+            bre.compile_search(r'\p{bad_bytes:y}', re.UNICODE)
 
         self.assertEqual(str(e.value), 'Invalid Unicode property!')
 
@@ -1950,13 +2168,11 @@ class TestExceptions(unittest.TestCase):
         text_pattern = r"(\w)+"
         pattern = re.compile(text_pattern)
 
-        with pytest.raises(ValueError) as excinfo:
-            bre.compile_replace(
-                pattern, r'{1[0o3f]}', bre.FORMAT
+        with pytest.raises(TypeError):
+            bre.subf(
+                pattern, r'{1[0o3f]}', 'test'
             )
 
-        assert "Capture index must be an integer!" in str(excinfo.value)
-
     def test_format_bad_capture_range(self):
         """Test a bad capture."""
 
@@ -1966,11 +2182,9 @@ class TestExceptions(unittest.TestCase):
             pattern, r'{1[37]}', bre.FORMAT
         )
 
-        with pytest.raises(IndexError) as excinfo:
+        with pytest.raises(IndexError):
             expand(pattern.match('text'))
 
-        assert "is out of range!" in str(excinfo.value)
-
     def test_require_compiled_pattern(self):
         """Test a bad capture."""
 
@@ -2171,13 +2385,12 @@ class TestConvenienceFunctions(unittest.TestCase):
     def test_fullmatch(self):
         """Test that `fullmatch` works."""
 
-        if PY34_PLUS:
-            m = bre.fullmatch(r'This is a test for match!', "This is a test for match!")
-            self.assertTrue(m is not None)
+        m = bre.fullmatch(r'This is a test for match!', "This is a test for match!")
+        self.assertTrue(m is not None)
 
-            p = bre.compile(r'This is a test for match!')
-            m = p.fullmatch("This is a test for match!")
-            self.assertTrue(m is not None)
+        p = bre.compile(r'This is a test for match!')
+        m = p.fullmatch("This is a test for match!")
+        self.assertTrue(m is not None)
 
     def test_search(self):
         """Test that `search` works."""
@@ -2361,3 +2574,55 @@ class TestConvenienceFunctions(unittest.TestCase):
                 p.sub(r'\ltest', 'tests')
         else:
             self.assertEqual(p.sub(r'\ltest', 'tests'), r'\ltest')
+
+
+class TestDeprecated(unittest.TestCase):
+    """Test deprecated."""
+
+    def test_lowercase(self):
+        """Test deprecated lower."""
+
+        bre.purge()
+        with warnings.catch_warnings(record=True) as w:
+            # Cause all warnings to always be triggered.
+            warnings.simplefilter("always")
+
+            bre.compile(r'\l')
+            self.assertTrue(len(w) == 1)
+            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))
+
+    def test_inverse_lowercase(self):
+        """Test inverse deprecated lower."""
+
+        bre.purge()
+        with warnings.catch_warnings(record=True) as w:
+            # Cause all warnings to always be triggered.
+            warnings.simplefilter("always")
+
+            bre.compile(r'\L')
+            self.assertTrue(len(w) == 1)
+            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))
+
+    def test_uppercase(self):
+        """Test deprecated upper."""
+
+        bre.purge()
+        with warnings.catch_warnings(record=True) as w:
+            # Cause all warnings to always be triggered.
+            warnings.simplefilter("always")
+
+            bre.compile(r'\c')
+            self.assertTrue(len(w) == 1)
+            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))
+
+    def test_inverse_uppercase(self):
+        """Test inverse deprecated upper."""
+
+        bre.purge()
+        with warnings.catch_warnings(record=True) as w:
+            # Cause all warnings to always be triggered.
+            warnings.simplefilter("always")
+
+            bre.compile(r'\C')
+            self.assertTrue(len(w) == 1)
+            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))
